! A Simplicistic Sokoban Surrogate
! Copyright (C) 2000,2003,2004 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.
!
! 2004-06-09	Added code to make 3S run under DOS.
! 2003-06-02	Added level set selection menu.
!		Added 'player on safe field' (%) to level file format.
! 		Added multi-line credits (C) records to level format.
!		Added "levels" entry to configuration.
!		Made (some) menus remember the most recent selection.
! 2003-05-30	Local level file re-named to "3s-levels".
!		Local level file now has precedence over global files.
!		Changed length field of path files to ASCII.
!		Made level file format more transparent; invisible
!		areas are now detected using floodfill.
! 2003-05-29	Implemented multiple level sets.
! 2003-01-18	Added Save Game and Load Game options.
! 2003-01-18	Added auto-detection of operating system
!		(required to avoid .filenames under DOS).

module triple_s(t3x, iostream, char, string, ttyctl, system, config);

object	t[t3x], tty[ttyctl], chr[char], str[string], sys[system];

const	Debug = 0;

const	CHR_PLAYER	= '@';
const	CHR_PLAYER_SAFE	= '%';
const	CHR_BOX		= '#';
const	CHR_SAFEBOX	= '!';
const	CHR_WALL	= 'X';
const	CHR_FLOOR	= '\s';
const	CHR_PLACE	= ':';

const	MAXSETS =	16;
const	MAXCOLS =	64;
const	MAXROWS =	22;
const	ALLOWEDCOLS =	MAXCOLS-4;
const	ALLOWEDROWS =	MAXROWS-2;
const	BUFLEN =	1024;
const	MAXLEVELS =	256;
const	MAXMOVES =	8192;
const	MAXNAMELEN =	12;

var	Set::128, SetFile::256;
var	Cwd::256;
var	LevelDir::256;
object	Levels[iostream];
var	LevelsOpen;
object	Conf[config];
var	Local;
var	Lbuf[BUFLEN];
var	Loffset[MAXLEVELS], Nlevels;

var	Screen::MAXCOLS*MAXROWS+1;
var	Credits::MAXCOLS*MAXROWS;
var	Greeting::81, GreetingDone;
var	Nrows, Ncols,
	Xoff, Yoff;
var	Ncredits;
var	Levno;
var	X, Y;
var	Moves, Pushes, Boxes, Safe;

var	Message, Prevmsg;

var	Path::MAXMOVES+6, Pp, Ptop, Prpt;

var	S3dir::256;

var	Usecolor, Usemaps, Showstats, Showhelp;
var	Map;
var	Levelmap::MAXLEVELS;
var	C_text,
	C_standout;

var	C_player, C_wall, C_floor, C_place, C_box, C_safe, C_void;

standout() ie (Usecolor)
		tty.color(C_standout);
	else
		tty.color(TTYCTL.F_BLACK|TTYCTL.B_GREY);

standend()
	ie (Usecolor)
		tty.color(C_text);
	else
		tty.color(TTYCTL.F_GREY|TTYCTL.B_BLACK);

quit(msg, args) do var b::256;
	sys.fini();
	tty.mode(0);
	tty.color(C_void);
	tty.clear();
	tty.reset();
	tty.fini();
	t.write(2, t.newline(b), str.length(b));
	t.write(2, "FATAL ERROR: ", 13);
	t.write(2, str.format(b, msg, args), str.length(b));
	t.write(2, t.newline(b), str.length(b));
	t.write(2, b, str.length(b));
	halt 5;
end

var msgbuf::100;
xmsg(m, args, wait) do var k;
	Message := m;
	if (m) str.format(msgbuf, m, args);
	if (Usecolor) standend();
	tty.move(0, 22);
	tty.clreol();
	if (\m) return;
	Message := msgbuf;
	tty.move(39-str.length(msgbuf)/2, 22);
	standout();
	tty.writes(msgbuf);
	standend();
	if (wait) do
		tty.move(35, 23);
		tty.writes("PRESS ESC");
		while (1) do
			k := tty.readc();
			if (k = TTYCTL.K_ESC \/ k = ':') leave;
		end
		xmsg(0, [0], 0);
		tty.move(0, 23);
		tty.clreol();
	end
end

msg(m, wait) xmsg(m, [0], wait);

init() do var dd;
	sys.init();
	tty.init();
	tty.clear();
	chr.init();
	tty.mode(1);
	C_void := TTYCTL.F_GREY | TTYCTL.B_BLACK;
	if (tty.lines() < 24 \/ tty.columns() < 80)
		quit("screen too small (must be at least 80x24)", 0);
	dd := sys.opendir("/usr/local/share/3s");
	ie (dd = %1) do
		str.copy(LevelDir, "levels");
	end
	else do
		str.copy(LevelDir, "/usr/local/share/3s");
		sys.closedir(dd);
	end
	str.copy(Set, "default");
	Message := 0;
	Prevmsg := 0;
	Usecolor := 1;
	Usemaps := 1;
	Showstats := 1;
	Showhelp := 1;
	LevelsOpen := 0;
	C_text := TTYCTL.F_GREY | TTYCTL.B_BLACK;
	C_standout := TTYCTL.F_CYAN | TTYCTL.B_BLACK | TTYCTL.F_BRIGHT;
	C_player := TTYCTL.F_YELLOW  | TTYCTL.B_BLUE   | TTYCTL.F_BRIGHT;
	C_wall   := TTYCTL.F_BLUE    | TTYCTL.B_YELLOW;
	C_floor  := TTYCTL.F_CYAN    | TTYCTL.B_BLUE;
	C_place  := TTYCTL.F_MAGENTA | TTYCTL.B_BLUE;
	C_box    := TTYCTL.F_CYAN    | TTYCTL.B_GREY   | TTYCTL.F_BRIGHT;
	C_safe   := TTYCTL.F_MAGENTA | TTYCTL.B_GREY   | TTYCTL.F_BRIGHT;
	C_void   := C_text;
	Map := [
	[ CHR_PLAYER,	'@',	(C_player),	"Player"	],
	[ CHR_WALL,	'X',	(C_wall),	"Wall"		],
	[ CHR_FLOOR,	'\s',	(C_floor),	"Floor"		],
	[ CHR_PLACE,	':',	(C_place),	"Destination"	],
	[ CHR_BOX,	'#',	(C_box),	"Box"		],
	[ CHR_SAFEBOX,	'#',	(C_safe),	"Safe box"	],
	[ '.',		'\s',	(C_void),	"Void"		]
	];
end

loadlmap() do var b::256, fd;
	t.memfill(Levelmap, '0', MAXLEVELS);
	if (Local) return;
	str.format(b, "%S/%S/map", [(S3dir, Set)]);
	fd := t.open(b, T3X.OREAD);
	if (fd < 0) return -1;
	if (t.read(fd, Levelmap, MAXLEVELS) \= MAXLEVELS) do
		t.memfill(Levelmap, '0', MAXLEVELS);
		msg("invalid level map (cleared)", 1);
	end
	t.close(fd);
end

loadset(setname) do
	var	off;
	var	line::81, k;
	var	n;
	var	lpath::256;

	ie (setname::0 = 0) do
		Local := 1;
		str.copy(SetFile, "3s-levels");
	end
	else do
		Local := 0;
		str.format(SetFile, "%s/%s", [(LevelDir, setname)]);
	end
	if (LevelsOpen) Levels.close();
	if (Levels.open(SetFile, Lbuf, BUFLEN, IOSTREAM.FREAD) = -1)
		quit("cannot open level file: \"%S\"", [(SetFile)]);
	LevelsOpen := 1;
	if (\Local) do
		str.format(lpath, "%s/%s", [(S3Dir, setname)]);
		if (sys.chdir(lpath) < 0 /\ sys.mkdir(lpath) < 0)
			quit("failed to create directory: \"%S\"", [(lpath)]);
		sys.chdir(Cwd);
		str.format(lpath, "%s/%s/paths", [(S3Dir, setname)]);
		if (sys.chdir(lpath) < 0 /\ sys.mkdir(lpath) < 0)
			quit("failed to create directory: \"%S\"", [(lpath)]);
		sys.chdir(Cwd);
	end
	n := 0;
	Nlevels := 0;
	Ncredits := 0;
	off := Levels.reads(line, 81);
	if (t.memcomp(line, "3S-LEVELS", 9))
		quit("not a 3S levels file: %S", [(SetFile)]);
	while (1) do
		k := Levels.reads(line, 81);
		if (Levels.eof()) leave;
		if (k /\ line::(k-1) = '\n') do
			line::(k-1) := 0;
			if (k>1 /\ line::(k-2) = '\r')
				line::(k-2) := 0;
		end
		if (line::0 = 'C') do
			if (Ncredits >= ALLOWEDROWS)
				quit("too many rows in credits of %s",
					[(SetFile)]);
			line::(MAXCOLS-1) := 0;
			str.copy(@Credits::(Ncredits*MAXCOLS), line);
			Ncredits := Ncredits+1;
		end
		if (str.length(line) > ALLOWEDCOLS /\ \chr.digit(line::0))
			quit("too many columns in level #%D", [(Nlevels)]);
		if (chr.digit(line::0)) do
			if (n >= ALLOWEDROWS)
				quit("too many rows in level #%D",
					[(Nlevels)]);
			n := 0;
			if (Nlevels >= MAXLEVELS)
				quit("too many levels in \"%s\"", ["levels"]);
			Loffset[Nlevels] := off;
			Nlevels := Nlevels+1;
		end
		off := off+k;
		n := n+1;
	end
	Levels.reset();
	Levno := 0;
end

savelmap(exit) do var b::256, fd, p, q;
	if (Local) return;
	str.format(b, "%S/%S/map", [(S3dir, Set)]);
	fd := t.open(b, T3X.OWRITE);
	p := exit-> @quit: @msg;
	q := exit-> 0: 1;
	if (fd < 0) do
		call p("could not update level map", [(q)]);
	end
	if (t.write(fd, Levelmap, MAXLEVELS) \= MAXLEVELS) do
		call p("could not write new level map", [(q)]);
	end
	t.close(fd);
end

saveconf() do
	var	cpath::256;

	str.format(cpath, "%S/config", [(S3dir)]);
	if (conf.save(cpath, [
		"# This is a 3S 2.0 config file",
		"#",
		0 ])
	)
		quit("error saving configuration", 0);
end

loadconf() do
	var	cpath::256;
	var	e;
	const	C = CONFIG.T_COLOR, N = CONFIG.T_NUMERIC, S = CONFIG.T_STRING;

	Conf.init([
	[ "level-directory",	S,	(LevelDir),	@LevelDir	],
	[ "levels",		S,	"default",	@Set		],
	[ "use-color",		N,	1,		@Usecolor	],
	[ "use-maps",		N,	1,		@Usemaps	],
	[ "show-stats",		N,	1,		@Showstats	],
	[ "show-help",		N,	1,		@Showhelp	],
	[ "player-color",	C,	(C_player,	@Map[0][2])	],
	[ "wall-color",		C,	(C_wall,	@Map[1][2])	],
	[ "floor-color",	C,	(C_floor,	@Map[2][2])	],
	[ "place-color",	C,	(C_place,	@Map[3][2])	],
	[ "box-color",		C,	(C_box,		@Map[4][2])	],
	[ "safe-box-color",	C,	(C_safe,	@Map[5][2])	],
	[ "text-color",		C,	(C_text,	@C_text		],
	[ "void-color",		C,	(C_void,	@Map[6][2])	],
	[ "standout-color",	C,	(C_standout,	@C_standout	],
	[ "player-char",	N,	(Map[0][1],	@Map[0][1])	],
	[ "wall-char",		N,	(Map[1][1],	@Map[1][1])	],
	[ "floor-char",		N,	(Map[2][1],	@Map[2][1])	],
	[ "place-char",		N,	(Map[3][1],	@Map[3][1])	],
	[ "box-char",		N,	(Map[4][1],	@Map[4][1])	],
	[ "safe-box-char",	N,	(Map[5][1],	@Map[5][1])	],
	0 ]);
	str.format(cpath, "%S/config", [(S3dir)]);
	if (Conf.load(cpath, @e) = -1)
		xmsg("error in config (line %d), using defaults", [(e)], 1);
end

fini() do
	saveconf();
	tty.color(TTYCTL.F_GREY | TTYCTL.B_BLACK);
	tty.clear();
	tty.mode(0);
	tty.reset();
	tty.fini();
	Levels.close();
	halt;
end

readc() do var k;
	k := tty.readc();
	if (Message) do
		Prevmsg := Message;
		msg(0, 0);
	end
	return k;
end

flood(px, py) do
	if (	px < 0 \/ py < 0 \/ px >= Ncols \/ py >= Nrows \/
		Screen::(py*MAXCOLS+px) \= '\s'
	)
		return;
	Screen::(py*MAXCOLS+px) := '.';
	flood(px-1, py);
	flood(px, py-1);
	flood(px+1, py);
	flood(px, py+1);
end

fixup() do var p, i, q;
	Nrows := Nrows+1;
	Ncols := Ncols+2;
	for (i=0, Nrows) do
		p := @Screen::(i*MAXCOLS);
		p::Ncols := 0;
	end
	flood(0, 0);
	Ncols := Ncols-2;
	for (i=0, Nrows) do
		p := @Screen::(i*MAXCOLS);
		q := @Screen::((i+1)*MAXCOLS+1);
		t.memcopy(p, q, Ncols);
		p::Ncols := 0;
	end
	Nrows := Nrows-2;
	Y := Y-1;
end

loadmap(src) do var k, p, i, r, onsafe;
	r := 0;
	Nrows := 1;
	Ncols := 0;
	k := send(src, iostream, reads(Screen, 81));
	Levno := str.strtonum(Screen, 10, 0)-1;
	i := 0;
	while (Screen::i /\ Screen::i \= '\s')
		i := i+1;
	while (Screen::i = '\s')
		i := i+1;
	str.copy(Greeting, @Screen::i);
	GreetingDone := 0;
	i := str.length(Greeting);
	ie (i > 1 /\ Greeting::(i-2) = '\r')
		Greeting::(i-2) := 0;
	else
		Greeting::(i-1) := 0;
	Boxes := 0;
	Safe := 0;
	t.memfill(Screen, CHR_FLOOR, MAXCOLS*MAXROWS);
	while(1) do
		p := @Screen::(Nrows*MAXCOLS+1);
		k := send(src, iostream, reads(p, MAXCOLS));
		if (send(src, iostream, eof())) leave;
		if (k >= 0) p::k := CHR_FLOOR;
		if (k > 0 /\ p::(k-1) = '\n') p::(k-1) := CHR_FLOOR;
		if (k > 1 /\ p::(k-2) = '\r') p::(k-2) := CHR_FLOOR;
		p::(MAXCOLS-2) := 0;
		if (chr.digit(p::0)) do
			r := str.strtonum(p, 10, 0);
			leave;
		end
		i := str.scan(p, CHR_PLAYER);
		onsafe := 0;
		if (i = -1) do
			i := str.scan(p, CHR_PLAYER_SAFE);
			onsafe := 1;
		end
		if (i \= -1) do
			X := i;
			Y := Nrows;
			p::i := onsafe-> CHR_PLACE: CHR_FLOOR;
			p::(i+1) := onsafe-> CHR_PLACE: CHR_FLOOR;
		end
		for (i=0, k, 2) do
			if (p::i = CHR_BOX \/ p::i = CHR_SAFEBOX) do
				if (p::i = CHR_SAFEBOX) Safe := Safe+1;
				Boxes := Boxes+1;
			end
		end
		if (k > Ncols+1) Ncols := k-1;
		Nrows := Nrows+1;
	end
	t.memfill(@Screen::(Nrows*MAXCOLS), CHR_FLOOR, MAXCOLS); 
	fixup();
	Xoff := MAXCOLS/2-Ncols/2;
	if (Xoff < 0) Xoff := 0;
	Yoff := MAXROWS/2-Nrows/2;
	if (Yoff < 0) Yoff := 0;
	Moves := 0;
	Pushes := 0;
	return r;
end

loadlevel(n, wait) do
	if (n >= Nlevels \/ n < 0) do
		xmsg("no such level: %D", [(n+1)], wait);
		return -1;
	end
	Levels.move(Loffset[n], IOSTREAM.SEEK_SET);
	loadmap(Levels);
	Levels.reset();
	Levno := n;
end

showlevel() do
	var	row, buf::81, buf2;
	var	i, j, l, c, oc, nc, k;

	standend();
	tty.clear();
	for (row=0, Nrows) do
		buf2 := @Screen::(row*MAXCOLS);
		t.memcopy(buf, buf2, 81);
		ie (Usemaps) do
			for (i=0, 7) str.xlate(buf, Map[i][0], i);
			for (i=0, 7) str.xlate(buf, i, Map[i][1]);
		end
		else do
			str.xlate(buf, Map[6][0], Map[6][1]);
		end
		tty.move(Xoff, Yoff+row);
		ie (\Usecolor) do
			tty.writes(buf);
		end
		else do
			oc := -1;
			k := str.length(buf);
			for (i=0, k) do
				j := i;
				while (buf2::i = buf2::j)
					j := j+1;
				c := buf::j;
				buf::j := 0;
				nc := oc;
				for (l=0, 7) do
					if (buf2::i = Map[l][0]) do
						nc := Map[l][2];
						leave;
					end
				end
				if (nc \= oc) tty.color(nc);
				oc := nc;
				tty.writes(@buf::i);
				buf::j := c;
				i := j-1;
			end
		end
	end
end

writed4(cx, cy, n) do var b::33;
	tty.move(cx, cy);
	tty.writes(str.format(b, "%4:\sD", [(n)]));
end

writed3(cx, cy, n) do var b::33;
	tty.move(cx, cy);
	tty.writes(str.format(b, "%3:\sD", [(n)]));
end

scr(cx, cy) return Screen::(cy*MAXCOLS+cx);

setscr(cx, cy, c) do var i;
	Screen::(cy*MAXCOLS+cx) := c;
	Screen::(cy*MAXCOLS+cx+1) := c;
	tty.move(cx+Xoff, cy+Yoff);
	for (i=0, 6) do
		if (Map[i][0] = c) do
			if (Usemaps) c := Map[i][1];
			if (Usecolor) tty.color(Map[i][2]);
			leave;
		end
	end
	tty.writec(c);
	tty.writec(c);
end

canup() ie (	scr(X, Y-1) = CHR_FLOOR \/ scr(X, Y-1) = CHR_PLACE \/
		(scr(X, Y-1) = CHR_BOX \/ scr(X, Y-1) = CHR_SAFEBOX) /\
		 (scr(X, Y-2) = CHR_FLOOR \/ scr(X, Y-2) = CHR_PLACE)
	)
		return 1;
	else
		return 0;

candown() ie (	scr(X, Y+1) = CHR_FLOOR \/ scr(X, Y+1) = CHR_PLACE \/
		(scr(X, Y+1) = CHR_BOX \/ scr(X, Y+1) = CHR_SAFEBOX) /\
		 (scr(X, Y+2) = CHR_FLOOR \/ scr(X, Y+2) = CHR_PLACE)
	)
		return 1;
	else
		return 0;

canleft() ie (	scr(X-2, Y) = CHR_FLOOR \/ scr(X-2, Y) = CHR_PLACE \/
		(scr(X-2, Y) = CHR_BOX \/ scr(X-2, Y) = CHR_SAFEBOX) /\
		 (scr(X-4, Y) = CHR_FLOOR \/ scr(X-4, Y) = CHR_PLACE)
	)
		return 1;
	else
		return 0;

canright() ie (	scr(X+2, Y) = CHR_FLOOR \/ scr(X+2, Y) = CHR_PLACE \/
		(scr(X+2, Y) = CHR_BOX \/ scr(X+2, Y) = CHR_SAFEBOX) /\
		 (scr(X+4, Y) = CHR_FLOOR \/ scr(X+4, Y) = CHR_PLACE)
	)
		return 1;
	else
		return 0;

save(old, new)
	ie (old = CHR_BOX /\ new = CHR_SAFEBOX)
		Safe := Safe+1;
	else if (old = CHR_SAFEBOX /\ new = CHR_BOX)
		Safe := Safe-1;
			
movebox(cx, cy, dx, dy) do var h;
	h := scr(cx, cy);
	setscr(dx, dy, scr(dx, dy) = CHR_FLOOR-> CHR_BOX: CHR_SAFEBOX);
	setscr(cx, cy, h = CHR_BOX-> CHR_FLOOR: CHR_PLACE);
	save(h, scr(dx, dy));
end

addmove(dir) do
	Moves := Moves+1;
	if (chr.upper(dir)) Pushes := Pushes+1;
	if (Pp > MAXMOVES) return -1;
	if (Pp >= MAXMOVES) do
		msg("too many moves -- no longer saving", 1);
		Pp := MAXMOVES+1;
		return -1;
	end
	ie (\Pp \/ Path::(Pp-1) = '9') do
		Path::Pp := dir; Pp := Pp+1;
	end
	else ie (chr.digit(Path::(Pp-1)) /\ Path::(Pp-2) = dir) do
		Path::(Pp-1) := Path::(Pp-1) + 1;
	end
	else ie (Path::(Pp-1) = dir) do
		Path::Pp := '1'; Pp := Pp+1;
	end
	else do
		Path::Pp := dir; Pp := Pp+1;
	end
end

moveback() do var m;
	m := chr.digit(Path::(Pp-1))-> Path::(Pp-2): Path::(Pp-1);
	if (chr.upper(m)) do
		ie (m = 'I') movebox(X, Y-1, X, Y);
		else ie (m = 'J') movebox(X-2, Y, X, Y);
		else ie (m = 'K') movebox(X, Y+1, X, Y);
		else if (m = 'L') movebox(X+2, Y, X, Y);
		Pushes := Pushes-1;
	end
	ie (m = 'i' \/ m = 'I') Y := Y+1;
	else ie (m = 'j' \/ m = 'J') X := X+2;
	else ie (m = 'k' \/ m = 'K') Y := Y-1;
	else if (m = 'l' \/ m = 'L') X := X-2;
	Moves := Moves-1;
end

moveforward() do var m;
	m := chr.digit(Path::(Pp-1))-> Path::(Pp-2): Path::(Pp-1);
	ie (m = 'i' \/ m = 'I') Y := Y-1;
	else ie (m = 'j' \/ m = 'J') X := X-2;
	else ie (m = 'k' \/ m = 'K') Y := Y+1;
	else if (m = 'l' \/ m = 'L') X := X+2;
	if (chr.upper(m)) do
		ie (m = 'I') movebox(X, Y, X, Y-1);
		else ie (m = 'J') movebox(X, Y, X-2, Y);
		else ie (m = 'K') movebox(X, Y, X, Y+1);
		else if (m = 'L') movebox(X, Y, X+2, Y);
		Pushes := Pushes+1;
	end
	Moves := Moves+1;
end

undomove() do var m;
	if (\Pp) do
		msg("no moves done yet", 0);
		return -1;
	end
	if (Pp > MAXMOVES) do
		msg("too many moves - cannot undo", 0);
		return -1;
	end
	moveback();
	ie (chr.digit(Path::(Pp-1))) do
		m := Path::(Pp-1);
		Path::(Pp-1) := m - 1;
		if (m < '2') Pp := Pp-1;
	end
	else do
		Pp := Pp-1;
	end
end

moveup() do
	if (\canup()) return -1;
	Y := Y-1;
	ie (scr(X, Y) = CHR_BOX \/ scr(X, Y) = CHR_SAFEBOX) do
		movebox(X, Y, X, Y-1);
		addmove('I');
	end
	else do
		addmove('i');
	end
end

moveleft() do
	if (\canleft()) return -1;
	X := X-2;
	ie (scr(X, Y) = CHR_BOX \/ scr(X, Y) = CHR_SAFEBOX) do
		movebox(X, Y, X-2, Y);
		addmove('J');
	end
	else do
		addmove('j');
	end
end

movedown() do
	if (\candown()) return -1;
	Y := Y+1;
	ie (scr(X, Y) = CHR_BOX \/ scr(X, Y) = CHR_SAFEBOX) do
		movebox(X, Y, X, Y+1);
		addmove('K');
	end
	else do
		addmove('k');
	end
end

moveright() do
	if (\canright()) return -1;
	X := X+2;
	ie (scr(X, Y) = CHR_BOX \/ scr(X, Y) = CHR_SAFEBOX) do
		movebox(X, Y, X+2, Y);
		addmove('L');
	end
	else do
		addmove('l');
	end
end

status(full) do
	if (\Showstats) return 0;
	if (full) do
		if (Usecolor) standend();
		tty.move(64, 2); tty.writes("Level");
		tty.move(64, 4); tty.writes(" ___ Boxes");
		tty.move(64, 5); tty.writes(" ___ Safe");
		tty.move(64, 6); tty.writes("____ Moves");
		tty.move(64, 7); tty.writes("____ Pushes");
		if (Usecolor) standout();
		writed4(70, 2, Levno+1);
		if (Usecolor) standend();
	end
	if (Usecolor) standout();
	writed3(65, 4, Boxes);
	writed3(65, 5, Safe);
	writed4(64, 6, Moves);
	writed4(64, 7, Pushes);
	if (Usecolor) standend();
end

minihelp() do
	if (\Showhelp) return;
	if (Usecolor) standend();
	tty.move(64,  9); tty.writes("   move:");
	tty.move(64, 10); tty.writes("    [i]");
	tty.move(64, 11); tty.writes("     |");
	tty.move(64, 12); tty.writes("[j]-[k]-[l]");
	tty.move(64, 14); tty.writes("[shift] fast");
	tty.move(64, 15); tty.writes("[u]     undo");
	tty.move(64, 16); tty.writes("[F1]    help");
	tty.move(64, 17); tty.writes("[ESC]   menu");
end

clrmenu(left) do var i;
	if (Usecolor) standend();
	for (i=0, 20) do
		tty.move(left, i);
		tty.clreol();
	end
end

menu(dfl, title, items) do var i, n, k, j;
	clrmenu(64);
	if (Usecolor) standout();
	tty.move(64, 2); tty.writes(title);
	if (Usecolor) standend();
	n := 0;
	while (items[n]) do
		tty.move(64, n+4);
		tty.writes(items[n]);
		n := n+1;
	end
	i := dfl;
	while (1) do
		standout();
		tty.move(64, i+4);
		tty.writes(items[i]);
		k := readc();
		standend();
		tty.move(64, i+4);
		tty.writes(items[i]);
		ie (k = TTYCTL.K_ESC \/ k = TTYCTL.K_LEFT \/ k = ':') do
			i := -1;
			leave;
		end
		else ie (k = TTYCTL.K_RIGHT \/ k = 0x0D) do
			leave;
		end
		else ie (k = TTYCTL.K_UP \/ k = 'P'-'@') do
			i := i-1;
			if (i < 0) i := n-1;
		end
		else ie (k = TTYCTL.K_DOWN \/ k = 'N'-'@') do
			i := i+1;
			if (i >= n) i := 0;
		end
		else do
			if (chr.lower(k)) k := chr.ucase(k);
			for (j=0, n)
				if (k = items[j]::0) do
					i := j;
					leave;
				end
			if (j<n) leave;
		end
	end
	n := i;
	clrmenu(64);
	return n;
end

setplayer() do
	tty.move(X+Xoff, Y+Yoff);
	if (Usecolor) tty.color(Map[0][2]);
	tty.writec(Map[0][1]);
	tty.writec(Map[0][1]);
	tty.move(X+Xoff, Y+Yoff);
end


setcolor(name, color) do var i, j, k;
	clrmenu(60);
	tty.move(62, 2);
	standout();
	tty.writes("SET COLOR OF");
	standend();
	for (j=0, 8) do
		tty.move(62, j+6);
		for (i=0, 16) do
			tty.color(j << 4 | i);
			tty.writec('X');
		end
	end
	i := color[0] & 0x0F;
	j := color[0] >> 4 & 7;
	while (1) do
		tty.color(C_standout);
		tty.move(i+62, j+6); tty.writec(':');
		tty.move(62, 4);
		tty.color(j<<4 | i);
		tty.writes(name);
		k := readc();
		tty.move(i+62, j+6); tty.writec('X');
		ie (k = TTYCTL.K_LEFT \/ k = 'B'-'@') do
			i := i-1;
			if (i<0) i := 15;
		end
		else ie (k = TTYCTL.K_RIGHT \/ k = 'F'-'@') do
			i := i+1;
			if (i>15) i := 0;
		end
		else ie (k = TTYCTL.K_UP \/ k = 'P'-'@') do
			j := j-1;
			if (j<0) j := 7;
		end
		else ie (k = TTYCTL.K_DOWN \/ k = 'N'-'@') do
			j := j+1;
			if (j>7) j := 0;
		end
		else ie (k = 0x0D) do
			color[0] := j<<4 | i;
			leave;
		end
		else if (k = TTYCTL.K_ESC \/ k = ':') do
			leave;
		end
	end
	clrmenu(60);
end

change_colors() do var i;
	i := 0;
	while (1) do
		i := menu(i, "SET COLOR OF", [
			"Player",
			"Wall",
			"Floor",
			"Destination",
			"Box",
			"Safe box",
			"",
			"Normal Text",
			"Standout Text",
			"",
			"[ DONE ]",
		0] );
		if (i < 0 \/ i = 10) leave;
		if (i >= 6 /\ i <= 9) do
			if (i = 7 \/ i = 8)
				setcolor(i=7-> "Normal Text": "Standout Text",
					i=7-> @C_text: @C_standout);
				Map[6][2] := C_text;
			loop;
		end
		setcolor(Map[i][3], @Map[i][2]);
		if (Usecolor) do
			showlevel();
			setplayer();
		end
	end
end

setmapping(name, pmap) do var i, j, k;
	clrmenu(60);
	tty.move(62, 2);
	standout();
	tty.writes("MAP OBJECT");
	tty.move(62, 4);
	tty.writes(name);
	standend();
	for (j=2, 8) do
		tty.move(62, j+4);
		for (i=0, 16) do
			tty.writec(j*16+i < '\s'-> '\s': j*16+i);
		end
	end
	j := pmap[0] / 16;
	i := pmap[0] mod 16;
	while (1) do
		tty.move(i+62, j+4);
		k := readc();
		ie (k = TTYCTL.K_LEFT \/ k = 'B'-'@') do
			i := i-1;
			if (i<0) i := 15;
		end
		else ie (k = TTYCTL.K_RIGHT \/ k = 'F'-'@') do
			i := i+1;
			if (i>15) i := 0;
		end
		else ie (k = TTYCTL.K_UP \/ k = 'P'-'@') do
			j := j-1;
			if (j<2) j := 7;
		end
		else ie (k = TTYCTL.K_DOWN \/ k = 'N'-'@') do
			j := j+1;
			if (j>7) j := 2;
		end
		else ie (k = 0x0D) do
			pmap[0] := j*16 | i;
			leave;
		end
		else if (k = TTYCTL.K_ESC \/ k = ':') do
			leave;
		end
	end
	clrmenu(60);
end

change_map() do var i;
	i := 0;
	while (1) do
		i := menu(i, "COLORS", [
			"Player",
			"Wall",
			"Floor",
			"Destination",
			"Box",
			"Safe box",
			"",
			"[ DONE ]",
		0] );
		if (i < 0 \/ i = 7) leave;
		if (i = 6) loop;
		setmapping(Map[i][3], @Map[i][1]);
		if (Usemaps) do
			showlevel();
			setplayer();
		end
	end
end

preferences() do var i, m, r;
	r := 1;
	m := [
	"Colors   [_]",
	"'-change",
	"Mappings [_]",
	"'-change",
	"Status   [_]",
	"Help     [_]",
	"",
	"[ DONE ]",
	0 ];
	i := 0;
	while (1) do
		m[0]::10 := Usecolor-> 'X': '\s';
		m[2]::10 := Usemaps-> 'X': '\s';
		m[4]::10 := Showstats-> 'X': '\s';
		m[5]::10 := Showhelp-> 'X': '\s';
		i := menu(i, "PREFERENCES", m);
		ie (i = 0) do
			Usecolor := \Usecolor;
			r := 2;
		end
		else ie (i = 1) do
			change_colors();
			r := 2;
		end
		else ie (i = 2) do
			Usemaps := \Usemaps;
			r := 2;
		end
		else ie (i = 3) do
			change_map();
			r := 2;
		end
		else ie (i = 4) do
			Showstats := \Showstats;
		end
		else ie (i = 5) do
			Showhelp := \Showhelp;
		end
		else if (i = 7 \/ i = -1) do
			return r;
		end
	end
end

wh(cx, cy, m) do var i, j, k, so;
	tty.move(cx, cy);
	standend();
	so := 0;
	k := str.length(m);
	if (k = 0) return;
	for (i=0, k) do
		j := i;
		while (m::i /\ m::i \= '$')
			i := i+1;
		if (\m::i) leave;
		m::i := 0;
		tty.writes(@m::j);
		m::i := '$';
		j := i;
		ie (so) do
			standend();
			so := 0;
		end
		else do
			standout();
			so := 1;
		end
	end
 	if (m::j \= '$') tty.writes(@m::j);
end

help() do
	if (Usecolor) standend();
	tty.clear();
	wh(2, 1, "$Key assignments$ ($[^X]$ denotes $[control]$ + $[X]$)");
	wh(2, 3, "$Keys used in the game");
	wh(2, 5, "     $[i]$       The keys [ijkl] move");
	wh(2, 6, "      $|$        the player figure up,");
	wh(2, 7, "$[j]--[k]--[l]$  down, left, and right.");
	wh(2, 8, "Holding down $[shift]$ simultaneously");
	wh(2, 9, "moves the figure until it blocks.");
	wh(2,11, "$[u]$    move back (undo)");
	wh(2,12, "$[^P]$   re-print last message");
	wh(2,13, "$[ESC]$  open main menu");
	wh(2,15, "$Keys used in menus");
	wh(2,17, "$[^N][^P]$  move down/up");
	wh(2,18, "$[ENTER]$   select and return");
	wh(2,19, "$[ESC]$     cancel and return");
	wh(2,20, "Menu items also can be activated by");
	wh(2,21, "pressing their first letters.");
	wh(2,23, "$press [ENTER] to return to the game");
	wh(42, 3, "$Keys used in the selection screen");
	wh(42, 5, "$[^N][^P]$  goto next/prev. level");
	wh(42, 6, "$[ENTER]$   play level");
	wh(42, 7, "$[ESC]$     return to opening menu");
	wh(42, 8, "$[t]$       trace (if available)");
	wh(42,10, "$Keys used in trace mode");
	wh(42,12, "$[j][k]$  trace forward/backward");
	wh(42,13, "$[J][K]$  trace until key change");
	wh(42,14, "$[^N]$    trace until box safe");
	wh(42,15, "$[^P]$    trace until box unsafe");
	wh(42,16, "$[ESC]$   exit trace mode");
	wh(42,18, "Mostly, the $[arrow keys]$ can be");
	wh(42,19, "used intead of $[ijkl]$ and $[^P][^N]");
	wh(42,20, "and $[:]$ is equal to $[ESC]$.");
	wh(42,21, "$[^L]$ causes a screen refresh.");
	readc();
	return 2;
end

confirm(what) return menu(0, what, [ "Yes", "No", 0 ]) = 0 -> 1: 0;

exitlevel() return confirm("EXIT LEVEL") = 0-> 1: -1;

doquit() return confirm("QUIT GAME");

restart() do
	if (\confirm("RESTART")) return 1;
	loadlevel(Levno, 0);
	Pp := 0;
	return 2;
end

trim(a, b) do var i;
	t.memcopy(b, a, MAXCOLS);
	b::(MAXCOLS+1) := 0;
	for (i=MAXCOLS, -1, -1) do
		if (b::i \= '\s') do
			b::(i+1) := 0;
			leave;
		end
	end
	return b;
end

savegame() do
	object	m[iostream];
	var	cbuf::256, tmp::256;
	var	p, i;

	str.format(tmp, "%S/saved", [(S3dir)]);
	if (m.open(tmp, cbuf, 256, IOSTREAM.FWRITE) = -1) do
		msg("Failed to save game", 0);
		return 1;
	end
	m.writes(str.format(tmp, "3S-SAVED %d% %d %s\n", [(X), (Y),
		(Local-> "-": Set)]));
	m.writes(str.format(tmp, "%d%s%s\n", [(Levno+1),
		(Greeting::0-> "\s": ""), (Greeting)]));
	for (i=0, Nrows) do
		p := @Screen::(i*MAXCOLS);
		m.writes(trim(p, tmp));
		m.writes(t.newline(tmp));
	end
	m.writes(str.format(tmp, "%d\n", [(Pp)]));
	m.writes(str.format(tmp, "%d %d\n", [(Moves), (Pushes)]));
	m.write(Path, Pp);
	m.writes(t.newline(tmp));
	m.close();
	msg("Saved.", 0);
	return 1;
end

loadgame(ok) do
	object	m[iostream];
	var	cbuf::256, tmp::256;
	var	p, i, nx, ny, bad;
	var	lset::128;

	str.format(tmp, "%S/saved", [(S3dir)]);
	if (m.open(tmp, cbuf, 256, IOSTREAM.FREAD) = -1) do
		msg("No game saved.", ok);
		return 1;
	end
	if (\ok /\ \confirm("LOAD GAME")) do
		m.close();
		return 1;
	end
	m.reads(tmp, 256);
	if (str.parse(tmp, "3S-SAVED %d %d %:\ns\n", [(@nx, @ny, lset)])
		\= 3
	) do
		msg("Saved map is invalid", ok);
		m.close();
		return 1;
	end
	ie (\str.comp(lset, "-")) do
		loadset("");
	end
	else do
		str.copy(Set, lset);
		loadset(Set);
		loadlmap();
	end
	Pp := loadmap(m);
	X := nx;
	Y := ny;
	bad := 0;
	m.reads(tmp, 256);
	if (str.parse(tmp, "%d %d", [(@Moves), (@Pushes)]) \= 2) bad := 1;
	if (m.read(Path, Pp) \= Pp) bad := 1;
	m.close();
	if (bad) do
		msg("Saved map is corrupt. Bailing out.", 1);
		fini();
	end
	return 2;
end

gamemenu() do var i;
	i := menu(0, "GAME MENU", [
		"Preferences",
		"Restart",
		"Save game",
		"Load game",
		"Exit level",
		"Help",
		"Quit",
	0 ]);
	ie (i = 0) return preferences();
	else ie (i = 1) return restart();
	else ie (i = 2) return savegame();
	else ie (i = 3) return loadgame(0);
	else ie (i = 4) return exitlevel();
	else ie (i = 5) return help();
	else if (i = 6) if (doquit()) fini();
	return 1;
end

savepath(n) do var b::256, fd, k, u, nb::7;
	if (Local) return;
	str.format(b, "%S/%S/paths/%3:0D", [(S3dir, Set, n+1)]);
	fd := t.open(b, T3X.OREAD);
	k := MAXMOVES+1;
	if (fd >= 0) do
		if (t.read(fd, nb, 5) > 1) do
			nb::5 := 0;
			k := str.strtonum(nb, 10, 0);
		end
		t.close(fd);
	end
	if (Moves < k) do
		fd := t.open(b, T3X.OWRITE);
		if (fd >= 0) do
			str.format(nb, "%d,", [(Moves)]);
			u := t.write(fd, nb, str.length(nb));
			u := u + t.write(fd, Path, Pp);
			t.close(fd);
			if (u \= Pp+str.length(nb)) do
				t.remove(b);
				msg("could not write path file", 0);
			end
		end
	end
end

loadpath(n) do
	var	b::256, fd, k;
	var	m;

	if (Local) return;
	m := "bad trace file: level %D";
	str.format(b, "%S/%S/paths/%3:0D", [(S3dir, Set, n+1)]);
	fd := t.open(b, T3X.OREAD);
	if (fd < 0) do
		msg("no trace information avaliable for level %D", [(n+1)]);
		return -1;
	end
	Pp := t.read(fd, Path, MAXMOVES+6);
	t.close(fd);
	k := t.memscan(Path, ',', Pp);
	if (k = -1) do
		xmsg(m, [(n+1)], 1);
		Pp := 0;
		return -1;
	end
	k := k+1;
	Pp := Pp-k;
	if (Pp >= MAXMOVES) do
		xmsg(m, [(n+1)], 1);
		Pp := 0;
		return -1;
	end
	t.memcopy(Path, @Path::k, Pp);
end

play() do var k, red, auto;
	red := 2;
	auto := 0;
	while (Safe < Boxes) do
		if (Debug) do var i;
			tty.move(0, 23);
			for (i=Pp-20, Pp)
				if (i>=0) tty.writec(Path::i);
			tty.writec('\s');
		end
		if (red) do
			if (red = 2) showlevel();
			minihelp();
			status(1);
			red := 0;
			if (\GreetingDone /\ Greeting::0) msg(Greeting, 0);
			GreetingDone := 1;
		end
		status(0);
		setplayer();
		if (Message) auto := 0;
		k := auto-> auto: readc();
		setscr(X, Y, scr(X, Y));
		ie (k = 'i' \/ k = TTYCTL.K_UP) do
			if (moveup()) auto := 0;
		end
		else ie (k = 'j' \/ k = TTYCTL.K_LEFT) do
			if (moveleft()) auto := 0;
		end
		else ie (k = 'k' \/ k = TTYCTL.K_DOWN) do
			if (movedown()) auto := 0;
		end
		else ie (k = 'l' \/ k = TTYCTL.K_RIGHT) do
			if (moveright()) auto := 0;
		end
		else ie (k = 'u') do
			undomove();
		end
		else ie (k = 'I') auto := 'i';
		else ie (k = 'J') auto := 'j';
		else ie (k = 'K') auto := 'k';
		else ie (k = 'L') auto := 'l';
		else ie (k = 'L'-'@' \/ k = 'R'-'@') red := 2;
		else ie (k = 'P'-'@') msg(Prevmsg, 0);
		else ie (k = TTYCTL.K_F1 \/ k = '?') do
			red := help();
		end
		else if (k = TTYCTL.K_ESC \/ k = ':') do
			red := gamemenu();
			if (red < 0) leave;
		end
	end
	if (Safe = Boxes) do
		savepath(Levno);
		setplayer();
		xmsg("%D moves", [(Moves)], 0);
		Levelmap::Levno := '1';
		savelmap(0);
	end
end

playlevel(n) do
	if (loadlevel(n, 1) < 0) return -1;
	Pp := 0;
	play();
end

atfirstmove() msg("already at first move", 0);

atlastmove() msg("already at last move", 0);

traceback() do
	if (Pp < 1) do
		atfirstmove();
		return -1;
	end
	moveback();
	ie (Prpt) do
		Prpt := Prpt-1;
	end
	else do
		ie (chr.digit(Path::(Pp-1))) do
			Pp := Pp-1;
			Prpt := Path::Pp-'1';
		end
		else do
			Pp := Pp-1;
		end
	end

end

traceforward() do
	if (Pp+1 > Ptop) do
		atlastmove();
		return -1;
	end
	Pp := Pp+1;
	if (Pp /\ chr.digit(Path::(Pp-1))) do
		Prpt := Prpt+1;
		ie (Prpt+'1' > Path::(Pp-1)) do
			Prpt := 0;
		end
		else do
			Pp := Pp-1;
		end
	end
	moveforward();
end

tracehelp() do
	if (\Showhelp) return;
	if (Usecolor) standend();
	tty.move(64,  9); tty.writes("[l] forward");
	tty.move(64, 10); tty.writes("[j] backward");
	tty.move(64, 11); tty.writes("[L] fast fwd");
	tty.move(64, 12); tty.writes("[J] fast bkw");
	tty.move(64, 13); tty.writes("[Next] save");
	tty.move(64, 14); tty.writes("[Prev] unsave");
	tty.move(64, 15); tty.writes("[ESC]  exit");
end

trace() do var red, k, old, new;
	Ptop := Pp;
	Pp := 0;
	Prpt := 0;
	red := 1;
	while (1) do
		if (red) do
			showlevel();
			tracehelp();
			if (Usecolor) standout();
			tty.move(64, 2); tty.writes("TRACE MODE");
			tty.move(64, 4); tty.writes("move");
			standend();
			red := 0;
		end
		if (Usecolor) standout();
		writed4(69, 4, Moves);
		if (Usecolor) standend();
		setplayer();
		k := readc();
		setscr(X, Y, scr(X, Y));
		ie (k = TTYCTL.K_ESC \/ k = ':') do
			leave;
		end
		else ie (k = TTYCTL.K_LEFT \/ k = 'j') do
			traceback();
		end
		else ie (k = TTYCTL.K_RIGHT \/ k = 'l') do
			traceforward();
		end
		else ie (k = 'J') do
			if (Pp < 1) atfirstmove();
			old := chr.lcase(chr.digit(Path::(Pp-1))->
				Path::(Pp-2): Path::(Pp-1));
			while (Pp > 0) do
				traceback();
				new := chr.lcase(chr.digit(Path::(Pp-1))->
					Path::(Pp-2): Path::(Pp-1));
				if (old \= new) leave;
			end
		end
		else ie (k = 'L') do
			if (Pp >= Ptop) atlastmove();
			old := chr.lcase(chr.digit(Path::Pp)->
				Path::(Pp-1): Path::Pp);
			while (Pp < Ptop) do
				traceforward();
				new := chr.lcase(chr.digit(Path::Pp)->
					Path::(Pp-1): Path::Pp);
				if (old \= new) leave;
			end
		end
		else ie (k = TTYCTL.K_NEXT \/ k = 'N'-'@') do
			if (Pp >= Ptop) atlastmove();
			old:= Safe;
			while (Safe = old /\ Pp < Ptop)
				traceforward();
		end
		else ie (k = TTYCTL.K_PREV \/ k = 'P'-'@') do
			traceback();
			old := Safe;
			while (Safe = old /\ Pp > 0)
				traceback();
		end
		else ie (k = 'L'-'@' \/ k = 'R'-'@') do
			red := 1;
		end
		else do
		end
	end
end

tracelevel(n) do
	if (loadlevel(n, 1) < 0) return -1;
	loadpath(n);
	trace();
end

getnum(cx, cy, p, k) do
	var	i, l;
	var	b::4;

	i := 0;
	tty.move(cx, cy);
	tty.writes(p);
	l := str.length(p);
	while (1) do
		k := k-> k: readc();
		ie (k = '\b') do
			if (\i) return -1;
			i := i-1;
			tty.move(cx+l+i, cy);
			tty.writec('\s');
			tty.move(cx+l+i, cy);
		end
		else ie (k = TTYCTL.K_ESC \/ k = ':') do
			return -1;
		end
		else ie (k = 0x0D) do
			b::i := 0;
			return str.strtonum(b, 10, @k);
		end
		else ie (chr.digit(k)) do
			if (i < 3) do
				tty.move(cx+l+i, cy);
				tty.writec(k);
				b::i := k;
				i := i+1;
			end
		end
		else do
		end
		k := 0;
	end
end

selectlevel() do
	var	k, fd, nb::7;
	var	b::256, n, c::2;
	var	red;

	red := 1;
	while (1) do
		ie (red) do
			loadlevel(Levno, 0);
			showlevel();
			setplayer();
			if (Greeting::0) msg(Greeting, 0);
			standout();
			tty.move(64, 2);
			tty.writes("SELECT LEVEL");
			tty.move(64, 4);
			tty.writes("Level");
			red := 0;
		end
		else do
			standout();
		end
		writed3(70, 4, Levno+1);
		standend();
		tty.move(64, 6); tty.writes("[down]  next");
		tty.move(64, 7); tty.writes("[up]    prev.");
		tty.move(64, 8); tty.writes("[0-9]   goto #");
		tty.move(64, 9); tty.writes("[ENTER] play");
		tty.move(64,10); tty.writes("[ESC]   return");
		str.format(b, "%S/%S/paths/%3:0D", [(S3dir, Set, Levno+1)]);
		fd := Local-> -1: t.open(b, T3X.OREAD);
		if (fd >= 0) do
			if (t.read(fd, nb, 5) > 1) do
				nb::5 := 0;
				n := str.strtonum(nb, 10, 0);
				standout();
				writed4(64, 12, n);
				tty.move(69, 12); tty.writes("moves");
				standend();
				tty.move(64, 14); tty.writes("[t] = trace");
			end
			t.close(fd);
		end
		k := readc();
		ie (k = TTYCTL.K_UP \/ k = TTYCTL.K_PREV \/ k = 'P'-'@') do
			Levno := Levno-1;
			if (Levno < 0) Levno := Nlevels-1;
			red := 1;
		end
		else ie (k = TTYCTL.K_DOWN \/ k = TTYCTL.K_NEXT \/
			k = 'N'-'@'
		) do
			Levno := Levno+1;
			if (Levno >= Nlevels) Levno := 0;
			red := 1;
		end
		else ie (k = TTYCTL.K_RIGHT \/ k = 0x0D) do
			playlevel(Levno);
			if (Safe = Boxes) xmsg("%D moves", [(Moves)], 1);
			red := 1;
		end
		else ie (k = TTYCTL.K_LEFT \/ k = TTYCTL.K_ESC \/ k = ':') do
			return;
		end
		else ie ((k = 't' \/ k = 'T') /\ fd >= 0) do
			tracelevel(Levno);
			red := 1;
		end
		else ie (k = 'L'-'@' \/ k = 'R'-'@') do
			red := 1;
		end
		else if (chr.digit(k)) do
			standout();
			n := getnum(64, 16, "goto ", k);
			standend();
			if (n < 1) loop;
			loadlevel(n-1, 1);
			red := 1;
		end
	end
end

alldone() do
	msg(
"Time for something more challenging? Use 'Select set' in the main menu",
	1);
end

hello() do
	wh(10,4,"A $S$implicistic $S$okoban $S$urrogate");
	tty.move(10, 6);
	tty.writes("Copyright (C) 2000,2003,2004 Nils M Holm");
	tty.move(10, 9);
	tty.writes("Based upon the original Sokoban game,");
	tty.move(10, 10);
	tty.writes("apparently (C) 1982 Thinking Rabbit");
	wh(10,13, "Use the [cursor keys] and [ENTER] to select");
	wh(10,14, "an option from the main menu or just press");
	wh(10,15, "$[ENTER]$ to play.");
end

license() do
	tty.color(C_text);
	tty.clear();
	tty.color(C_standout);
	tty.move( 2, 1); tty.writes(
"A Simplicistic Sokoban Surrogate");
	tty.move( 2, 2); tty.writes(
"Copyright (C) 2000,2003,2004 Nils M Holm.  All rights reserved.");
	tty.color(C_text);
	tty.move( 2, 4); tty.writes("Don't worry, be happy.");
	tty.move( 2,6); tty.writes(
"THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND");
	tty.move( 2,7); tty.writes(
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE");
	tty.move( 2,8); tty.writes(
"IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE");
	tty.move( 2,9); tty.writes(
"ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE");
	tty.move( 2,10); tty.writes(
"FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL");
	tty.move( 2,11); tty.writes(
"DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS");
	tty.move( 2,12); tty.writes(
"OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)");
	tty.move( 2,13); tty.writes(
"HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT");
	tty.move( 2,14); tty.writes(
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY");
	tty.move( 2,15); tty.writes(
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF");
	tty.move( 2,16); tty.writes(
"SUCH DAMAGE.");
	tty.move(80-14, 24-1);
	tty.color(C_standout);
	tty.writes("Press [ENTER]");
	tty.color(C_text);
	tty.readc();
end

exists(file) do var fd;
	fd := t.open(file, T3X.OREAD);
	if (fd < 0) return 0;
	t.close(fd);
	return 1;
end

showCredits() do var i, b;
	tty.color(C_text);
	tty.clear();
	b := 11 - Ncredits/2;
	for (i=0, Ncredits) wh(2, b+i, @Credits::(i*MAXCOLS+1));
end

selectset() do
	var	sets[MAXSETS+3];
	var	names::MAXSETS+3*MAXNAMELEN, nt;
	var	i, dir, k, n;
	var	haveLocal;
	var	lset::128;

	standend();
	tty.clear();
	n := 0;
	nt := 0;
	haveLocal := 0;
	if (exists("3s-levels")) do
		sets[n] := "Local Levels";
		n := n+1;
		haveLocal := 1;
	end
	dir := sys.opendir(LevelDir);
	i := 0;
	if (dir >= 0) do
		while (1) do
			k := sys.readdir(dir, @names::nt, MAXNAMELEN+2);
			if (k < 1) leave;
			ie (k > MAXNAMELEN) do
				xmsg("level name too long: %S",
					[(@names::nt)], 1);
			end
			else ie (n >= MAXSETS) do
				xmsg("too many level files", 0, 1);
				leave;
			end
			else ie (names::nt = '.') do
				! ignore
			end
			else do
				if (\str.comp(@names::nt, Set)) i := n;
				sets[n] := @names::nt;
				nt := nt+k+1;
				n := n+1;
			end
		end
		sys.closedir(dir);
	end
	sets[n] := "";
	sets[n+1] := "[ SELECT ]";
	sets[n+2] := 0;
	n := n+1;
	showCredits();
	str.copy(lset, Set);
	while (1) do
		i := menu(i, "LEVEL SETS", sets);
		ie (i = -1) do
			loadset(Set);
			loadlmap();
			leave;
		end
		else ie (i = n-1) do
			! ignore
		end
		else ie (i = n) do
			str.copy(Set, lset);
			loadlmap();
			leave;
		end
		else do
			str.copy(lset, haveLocal /\ i = 0-> "": sets[i]);
			loadset(lset);
			showCredits();
		end
	end
end

mainmenu(i) do var j;
	standend();
	tty.clear();
	hello();
	i := menu(i, "MAIN MENU", [
		"Next level",
		"Load game",
		"Goto level",
		"Select set",
		"Preferences",
		"License",
		"Quit",
	0 ]);
	ie (i = 0) do
		j := t.memscan(Levelmap, '0', MAXLEVELS);
		while (1) do
			if (j < 0 \/ j > Nlevels-1) do
				alldone();
				return;
			end
			playlevel(j);
			if (Safe \= Boxes) leave;
			if (menu(0, "LEVEL DONE", [
				"Next level",
				"Main menu",
			0] ) \= 0)
				leave;
			j := t.memscan(Levelmap, '0', MAXLEVELS);
		end
	end
	else ie (i = 1) do
		if (loadgame(1) > 1) play();
	end
	else ie (i = 2) do
		selectlevel();
	end
	else ie (i = 3) do
		selectset();
	end
	else ie (i = 4) do
		loadlevel(0, 0);
		preferences();
	end
	else ie (i = 5) do
		license();
	end
	else if (i = 6 \/ i = -1) do
		fini();
	end
	return i;
end

signon() do
	var	home::240, tmp::10;
	var	fd;

	ie (t.getenv("COMSPEC", tmp, 9) > 0) do
		str.copy(S3dir, ".");
	end
	else do
		if (t.getenv("HOME", home, 240) = -1) str.copy(home, ".");
		str.format(S3dir, "%S/.3s", [(home)]);
	end
	if (sys.getdir(Cwd, 240) = -1)
		quit("Internal error: failed to get CWD:", 0);
	if (sys.chdir(S3dir) < 0 /\ sys.mkdir(S3dir) < 0)
		quit("failed to create directory: %S", [(S3Dir)]);
	sys.chdir(Cwd);
	loadconf();
	loadset(Set);
	loadlmap();
end

do var i;
	init();
	signon();
	i := 0;
	while (1) i := mainmenu(i);
end
