! Convert original Sokoban levels to own format.
! Copyright (C) 1991,2003 Nils M Holm.
!
! Usage: copy IMPORT.T to SOKOBAN directory and run:
!
! TX IMPORT.T
! TXX IMPORT
!
! IMPORT will then create a file called ORIGINAL that
! contains the original Sokoban levels in 3S format.
!
! To play the levels, copy the file ORIGINAL to the
! LEVELS directory of 3S and select the set from the
! main menu.

module sokoban_import(t3x, string, iostream, util, system);

object	t[t3x], str[string], u[util], sys[system];

const	BUFLEN	= 1024;

object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Offsets[100];

convert(name) do
	var	on1::3, on2::3;
	var	oname::15;
	var	offset;
	var	pos, xpos, left;
	var	c, pass;
	var	nl::4;

	t.newline(nl);
	u.printf("converting: %s", [(name)]);
	if (In.open(name, Inbuf, BUFLEN, IOSTREAM.FREAD) = -1) do
		u.writef(T3X.SYSERR, "%scannot open: %s%s", [(nl,name,nl)]);
		return;
	end
	ie (\t.memcomp(name, "TABLE", 5)) do
		on1::0 := '0';
		on1::1 := name::5;
		on1::2 := 0;
		on2::0 := '0';
		on2::1 := name::6;
		on2::2 := 0;
	end
	else ie (\t.memcomp(name, "TABL", 4)) do
		ie (name::4 = '9') do
			str.copy(on1, "09");
			str.copy(on2, "10");
		end
		else do
			on1::0 := '0';
			on1::1 := name::5;
			on1::2 := 0;
			on2::0 := '0';
			on2::1 := name::6;
			on2::2 := 0;
		end
	end
	else ie (\t.memcomp(name, "TAB", 3)) do
		on1::0 := name::3;
		on1::1 := name::4;
		on1::2 := 0;
		on2::0 := name::5;
		on2::1 := name::6;
		on2::2 := 0;
	end
	else do
		In.close();
		u.writef(T3X.SYSERR, "%sinvalid name: %s%s", [(nl,name,nl)]);
		return;
	end
	str.format(oname, "SOKOTMP.%s", [(on1)]);
	pass := 1;
	while (pass <= 2) do
		if (out.open(oname, Outbuf, BUFLEN, IOSTREAM.FWRITE) = -1) do
			u.writef(T3X.SYSERR, "%scannot create: %s%s",
				[(nl,oname,nl)]);
			return;
		end
		offset := 0;
		pos := In.rdch()+1;
		if (pos = 1) pos := 257;
		xpos := 0;
		left := 20;
		while (offset<305) do
			if (offset = pos) do
				c := In.rdch();
				out.wrch('*');
				offset := offset+1;
				xpos := xpos+1;
				loop;
			end
			c := In.rdch();
			ie (offset = 0) out.wrch('.');
			else ie (c = 0) out.wrch('.');
			else ie (c = 1) out.wrch('X');
			else ie (c = 3) out.wrch(':');
			else ie (c = 20) out.wrch('#');
			else if (c = 23) out.wrch('!');
			if (c /\ xpos < left) left := xpos;
			xpos := xpos+1;
			if ((offset mod 19) = 0) do
				out.wrch('\n');
				xpos := 0;
			end
			offset := offset+1;
		end
		out.close();
		Offsets[str.strtonum(@oname::8, 10, 0)] := left;
		str.format(oname, "SOKOTMP.%s", [(on2)]);
		pass := pass+1;
	end
	In.close();
	u.printf(" into %s and %s%s", [(on1, on2, nl)]);
end

allspace(s) do var i;
	i := 0;
	while (s::i = '\s') i := i+1;
	return s::i = '\n';
end

trim(s) do var i, inside;
	if (s::1 = '\n') return;
	i := str.length(s)-2;
	while (s::i = '.') i := i-1;
	s::(i+1) := '\n';
	s::(i+2) := 0;
	i := 0;
	inside := 0;
	while (s::i) do
		ie (s::i = '.')
			s::i := '\s';
		else 
			inside := 1;
		if (inside /\ s::i = '.') s::i := '\s';
		i := i+1;
	end
end

double(s) do var q::42, i, j;
	i := 0;
	j := 0;
	while (s::i) do
		q::j := s::i;
		q::(j+1) := s::i;
		i := i+1;
		j := j+2;
	end
	q::(j-1) := 0;
	str.copy(s, q);
	return s;
end

import(id) do
	var	line::42;
	var	name::14;
	var	i;
	var	nl::4;

	str.format(name, "SOKOTMP.%2:0d", [(id)]);
	t.newline(nl);
	if (In.open(name, Inbuf, BUFLEN, IOSTREAM.FREAD) = -1) do
		u.writef(T3X.SYSERR, "cannot open: %s%s", [(name,nl)]);
		return;
	end
	u.printf("importing: %s", [(name)]);
	In.reads(line, 21);
	while (\in.eof()) do
		trim(line);
		ie (line::1 = '\n') do
			str.format(line, "%d\n", [(id)]);
			Out.writes(line);
		end
		else if (\allspace(line)) do
			Out.writes(double(@line::Offsets[id]));
		end
		In.reads(line, 21);
	end
	In.close();
	!t.remove(name);
	u.printf(" - Done%s", [(nl)]);
end

do
	var	dd;
	var	last;
	var	name::14;
	var	nl::4;
	var	i;

	t.newline(nl);
	dd := sys.opendir(".");
	if (dd = -1) do
		u.writef(T3X.SYSERR, "cannot open cwd%s", [(nl)]);
		halt 1;
	end
	last := sys.readdir(dd, name, 14) < 0;
	while (\last) do
		if (\t.memcomp(name, "TAB", 3)) convert(name);
		last := sys.readdir(dd, name, 14) < 0;
	end
	sys.closedir(dd);
	if (Out.open("original", Outbuf, BUFLEN, IOSTREAM.FWRITE) = -1) do
		u.writef(T3X.SYSERR, "cannot create: original%s", [(nl)]);
		halt 1;
	end
	for (i=1, 51) import(i);
	Out.close();
end
